#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from .defs import *
from .scalars import cvxpy_obj
from .arrays import cvxpy_array
from .arrays import cvxpy_matrix
from scipy.linalg import sqrtm as sci_sqrtm
from functools import reduce

# Names
__all__ = ["hstack","vstack","sum","randn","rand",
           "seed","eye","zeros","ones","diag","sqrtm", "trace",
           "reshape"]

# Utility function: hstack
def hstack(seq):
    """
    | Stacks objects horizontally.

    :param seq: tuple or list of numbers,
                :ref:`scalar objects<scalar_ref>` or
                :ref:`multidimensional objects<multi_ref>`.
    :return: :ref:`array<array_obj>` or
             :ref:`matrix<matrix_obj>`.
    """

    # Check input
    if (type(seq) is not tuple and
        type(seq) is not list):
        raise TypeError('Invalid argument')

    # Process input
    new_list = []
    numeric = True
    for x in seq:
        if np.isscalar(x):
            new_list += [matrix(x)]
        elif type(x) is cvxpy_obj:
            new_list += [matrix(x.value)]
        elif type(x).__name__ in SCALAR_OBJS:
            numeric = False
            new_x = cvxpy_array(1,1)
            new_x[0,0] = x
            new_list += [new_x]
        elif type(x).__name__ in ARRAY_OBJS:
            numeric = False
            new_list += [x]
        elif type(x) is cvxpy_matrix:
            new_list += [x]
        else:
            raise TypeError('Invalid input')

    # Input is numeric
    if numeric:
        return np.hstack(new_list)

    # Verify dimensions
    m = new_list[0].shape[0]
    for x in new_list:
        if x.shape[0] != m:
            raise ValueError('Invalid dimensions')

    # Allocate new array
    n = int(np.sum([x.shape[1] for x in new_list]))
    new_ar = cvxpy_array(m,n)

    # Fill new array
    k = 0
    for x in new_list:
        for i in range(0,x.shape[0],1):
            for j in range(0,x.shape[1],1):
                new_ar[i,k+j] = x[i,j]
        k = k + x.shape[1]

    # Return new array
    return new_ar

# Utility function: vstack
def vstack(seq):
    """
    | Stacks objects vertically.

    :param seq: tuple or list of numbers,
                :ref:`scalar objects<scalar_ref>` or
                :ref:`multidimensional objects<multi_ref>`.
    :return: :ref:`array<array_obj>` or
             :ref:`matrix<matrix_obj>`.
    """

    # Check input
    if (type(seq) is not tuple and
        type(seq) is not list):
        raise TypeError('Invalid argument')

    # Process input
    new_list = []
    numeric = True
    for x in seq:
        if np.isscalar(x):
            new_list += [matrix(x)]
        elif type(x) is cvxpy_obj:
            new_list += [matrix(x.value)]
        elif type(x).__name__ in SCALAR_OBJS:
            numeric = False
            new_x = cvxpy_array(1,1)
            new_x[0,0] = x
            new_list += [new_x]
        elif type(x).__name__ in ARRAY_OBJS:
            numeric = False
            new_list += [x]
        elif type(x) is cvxpy_matrix:
            new_list += [x]
        else:
            raise TypeError('Invalid Input')

    # Input is numeric
    if numeric:
        return np.vstack(new_list)

    # Verify dimensions
    n = new_list[0].shape[1]
    for x in new_list:
        if x.shape[1] != n:
            raise ValueError('Invalid Dimensions')

    # Allocate new array
    m = int(np.sum([x.shape[0] for x in new_list]))
    new_ar = cvxpy_array(m,n)

    # Fill new array
    k = 0
    for x in new_list:
        for i in range(0,x.shape[0],1):
            for j in range(0,x.shape[1],1):
                new_ar[i+k,j] = x[i,j]
        k = k + x.shape[0]

    # Return new array
    return new_ar

# Utility function: sum
def sum(seq):
    """
    | Sums objects in sequence or entries of array-like object.

    :param seq: number,
                :ref:`scalar object<scalar_ref>`,
                :ref:`multidimensional object<multi_ref>`,
                tuple or list of such objects.
    :return: number,
             :ref:`scalar object<scalar_ref>` or
             :ref:`multidimensional object<multi_ref>`.
    """

    # Number or scalars object
    if (np.isscalar(seq) or
        type(seq).__name__ in SCALAR_OBJS):
        return seq

    # Matrix or object array
    elif (type(seq) is cvxpy_matrix or
          type(seq).__name__ in ARRAY_OBJS):
        return reduce(lambda x,y: x + y,
                      [seq[i,j] for i in range(0,seq.shape[0])
                       for j in range(0,seq.shape[1])],0)

    # List or tuple
    elif type(seq) is tuple or type(seq) is list:
        return reduce(lambda x,y: x + y,seq,0)

    # Other
    else:
        raise TypeError('Invalid argument')

# Utility function: randn
def randn(m,n):
    """
    | See **numpy.random.randn**.

    :param m: rows.
    :param n: columns.
    :return: :ref:`matrix<matrix_obj>`.
    """
    return matrix(np.random.randn(m,n))

# Utility function: rand
def rand(m,n):
    """
    | See **numpy.random.rand**.

    :param m: rows.
    :param n: columns.
    :return: :ref:`matrix<matrix_obj>`.
    """
    return matrix(np.random.rand(m,n))

# Utility function: seed
def seed(s=None):
    """
    | See **numpy.random.seed**.

    :param s: integer, array-like or ``None``.
    :return: ``NoneType``.
    """
    np.random.seed(s)

# Utility function: eye
def eye(n):
    """
    | See **numpy.eye**.

    :param n: rows.
    :return: :ref:`matrix<matrix_obj>`.
    """
    return matrix(np.eye(n))

# Utility function: zeros
def zeros(shape):
    """
    | See **numpy.zeros**.

    :param shape: integer or
                  pair of integers.
    :return: :ref:`matrix<matrix_obj>`.
    """
    return matrix(np.zeros(shape))

# Utility function: ones
def ones(shape):
    """
    | See **numpy.ones**.

    :param shape: integer or
                  pair of integers.
    :return: :ref:`matrix<matrix_obj>`.
    """
    return matrix(np.ones(shape))

# Utility function: diag
def diag(v):
    """
    | Extracts diagonal or constructs a diagonal array.

    :param v: number,
              :ref:`scalar object<scalar_ref>`,
              square or one dimensional
              :ref:`multidimensional object<multi_ref>`.
    :return: number,
             :ref:`scalar object<scalar_ref>` or
             :ref:`multidimensional object<multi_ref>`.
    """

    # Check input
    if (np.isscalar(v) or
        type(v).__name__ in SCALAR_OBJS):
        return v
    elif (type(v) is cvxpy_matrix or
          type(v).__name__ in ARRAY_OBJS):
        if v.shape[1]==1:
            v = v.T
    else:
        raise TypeError('Invalid argument type')

    # Get shape
    (m,n) = v.shape

    # Input is 2D
    if m!=1 and n!=1:

        # Must be square
        if m!=n:
            raise ValueError('Invalid dimensions')

        # cvxpy matrix
        if type(v) is cvxpy_matrix:
            return matrix(np.diag(v)).T

        # Object array
        else:
            new_ar = cvxpy_array(m,1)
            for i in range(0,m,1):
                new_ar[i,0] = v[i,i]
            return new_ar

    # Input is 1D
    else:

        # cvxpy matrix
        if type(v) is cvxpy_matrix:
            return matrix(np.diagflat(v))

        # Object array
        else:
            new_ar = cvxpy_array(n,n)
            for i in range(0,n,1):
                new_ar[i,i] = v[0,i]
            return new_ar

# Utility function: sqrtm
def sqrtm(x):
    """
    | See **scipy.linalg.sqrtm**.

    :param x: positive semidefinite matrix.
    :return: :ref:`matrix<matrix_obj>`.
    """
    return matrix(np.real(sci_sqrtm(x)))

# Utility function: trace
def trace(x):
    """
    | Computes trace.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              square :ref:`multidimensional object<multi_ref>`.
    :return: number or
             :ref:`scalar object<scalar_ref>`.
    """

    # Check type
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        return x
    elif ((type(x) is not cvxpy_matrix) and
          (type(x).__name__ not in ARRAY_OBJS)):
        raise TypeError('Invalid argument type')

    # Check shape
    if x.shape[0] != x.shape[1]:
        raise ValueError('Invalid dimensions')

    # Get trace
    return sum(diag(x))

# Utility function: reshape
def reshape(v, newshape):
    """
    | Reshapes the array to dimensions newshape (see np.reshape)
    | in FORTRAN (column-major) order.

    :param v: :ref:`array<array_obj>` or
              :ref:`matrix<matrix_obj>`.
    :param newshape: tuple with two integers.
    :return: the reshaped variable or matrix,
             :ref:`array<array_obj>` or
             :ref:`matrix<matrix_obj>`.
    """

    # Check input
    if (type(v) is cvxpy_matrix or
          type(v).__name__ in ARRAY_OBJS):
        pass
    else:
        raise TypeError('Invalid argument type')

    if not (hasattr(newshape, '__iter__') and
            len(newshape) == 2):
        raise TypeError('Invalid argument for newshape')

    # Get shape
    (m,n) = v.shape

    # Get new shape
    (mn,nn) = newshape

    if mn*nn != m*n:
        raise ValueError('Output dimension size does not match input dimension')

    # cvxpy matrix
    if type(v) is cvxpy_matrix:
        return matrix(np.reshape(v, newshape, order='F'))
    # Object array
    else:
        new_ar = cvxpy_array(mn,nn)
        for j in range(0,n,1):
            for i in range(0,m,1):
                k = j*m + i
                new_ar[k % mn,k / mn] = v[i,j]
        return new_ar

# Load modules
from .interface import matrix



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

# Names
BASE_VAR = 'v'
BASE_PARAM = 'a'

# Epsilon
EPSILON = 1e-12

# Variable structure
SYMMETRIC = 'symmetric'
LOWER_TRIANGULAR = 'lower_triangular'
UPPER_TRIANGULAR = 'upper_triangular'

# Parameter attribute
NONNEGATIVE = 'nonnegative'
NONPOSITIVE = 'nonpositive'

# Groups
VARIABLE_OBJS = ['cvxpy_scalar_var','cvxpy_var']
PARAMETER_OBJS = ['cvxpy_scalar_param','cvxpy_param']
SCALAR_OBJS = ['cvxpy_scalar_var','cvxpy_scalar_param','cvxpy_tree']
ARRAY_OBJS = ['cvxpy_var','cvxpy_param','cvxpy_array']
MATRIX_OBJS = ['cvxpy_matrix']
FORMAL_OBJS = VARIABLE_OBJS + PARAMETER_OBJS

# Object types
SET = 'set'
TREE = 'tree'
ARRAY = 'array'
OPERATOR = 'operator'
CONSTANT = 'constant'
FUNCTION = 'function'
VARIABLE = 'variable'
PARAMETER = 'parameter'
SPARRAY = 'sparray'

# Operations
MULTIPLICATION = '*'
SUMMATION = '+'
SUBTRACTION = '-'
LEFT_ADD = 'la'
LEFT_SUBTRACT = 'ls'
RIGHT_ADD = 'ra'
RIGHT_SUBTRACT = 'rs'
LEFT_MULTIPLY = 'lm'
RIGHT_MULTIPLY = 'rm'

# Expansion types
SDC = 'sdc'
SOC = 'soc'
DIF = 'dif'
PM = 'pm'

# Actions
MINIMIZE = 'minimize'
MAXIMIZE = 'maximize'

# Constraint types
EQUALS = '=='
LESS_EQUALS = '<='
GREATER_EQUALS = '>='
BELONGS = 'in'

# Configuration
CONFIGURATION ={'maxiters':100,
                'abstol':1e-7,
                'reltol':1e-6,
                'feastol':1e-6}

# Solver fields
OPTIMAL = 'optimal'
UNKNOWN = 'unknown'
DUAL_INFEASIBLE = 'dual infeasible'
PRIMAL_INFEASIBLE = 'primal infeasible'
PRIMAL_OBJECTIVE = 'primal objective'



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np

from pydsm.delsig import synthesizeNTF, synthesizeChebyshevNTF, clans
from pydsm.relab import cplxpair

__all__ = ["TestSynthesizeNTF", "TestSynthesizeChebyshevNTF",
           "TestClans"]


class TestSynthesizeNTF:
    def setUp(self):
        pass

    def test_default(self):
        z, p, k = synthesizeNTF()
        e_k = 1
        e_z = np.ones(3)
        e_p = [0.7654 - 0.2793j, 0.7654 + 0.2793j, 0.6694]
        e_p = cplxpair(e_p)
        np.testing.assert_almost_equal(k, e_k, 6)
        np.testing.assert_almost_equal(z, e_z, 6)
        np.testing.assert_almost_equal(p, e_p, 4)

    def test_BP6(self):
        z, p, k = synthesizeNTF(order=6, f0=0.3)
        e_k = 1
        e_z = np.concatenate(((-0.3090 + 0.9511j)*np.ones(3),
                              (-0.3090 - 0.9511j)*np.ones(3)))
        e_z = cplxpair(e_z)
        e_p = [-0.4238 - 0.7906j, -0.4238 + 0.7906j, -0.2480 - 0.7802j,
               -0.2480 + 0.7802j, -0.1290 - 0.8985j, -0.1290 + 0.8985j]
        e_p = cplxpair(e_p)
        np.testing.assert_almost_equal(k, e_k, 6)
        np.testing.assert_almost_equal(z, e_z, 4)
        np.testing.assert_almost_equal(p, e_p, 4)

    def test_opt1(self):
        z, p, k = synthesizeNTF(opt=1)
        e_k = 1
        e_z = [1.0000, 0.9993 - 0.0380j, 0.9993 + 0.0380j]
        e_z = cplxpair(e_z)
        e_p = [0.6692, 0.7652 - 0.2795j, 0.7652 + 0.2795j]
        e_p = cplxpair(e_p)
        np.testing.assert_almost_equal(k, e_k, 6)
        np.testing.assert_almost_equal(z, e_z, 4)
        np.testing.assert_almost_equal(p, e_p, 4)

    def test_opt3(self):
        z, p, k = synthesizeNTF(opt=3)
        e_k = 1
        e_z = [1.0000, 0.9993 - 0.0382j, 0.9993 + 0.0382j]
        e_z = cplxpair(e_z)
        e_p = [0.6692, 0.7652 - 0.2795j, 0.7652 + 0.2795j]
        e_p = cplxpair(e_p)
        np.testing.assert_almost_equal(k, e_k, 6)
        np.testing.assert_almost_equal(z, e_z, 4)
        np.testing.assert_almost_equal(p, e_p, 4)

    def test_given_zeroz(self):
        g_z = [1.0000, 0.9993 - 0.0380j, 0.9993 + 0.0380j]
        z, p, k = synthesizeNTF(opt=g_z)
        e_k = 1
        e_z = [1.0000, 0.9993 - 0.0380j, 0.9993 + 0.0380j]
        e_z = cplxpair(e_z)
        e_p = [0.6692, 0.7652 - 0.2795j, 0.7652 + 0.2795j]
        e_p = cplxpair(e_p)
        np.testing.assert_almost_equal(k, e_k, 6)
        np.testing.assert_almost_equal(z, e_z, 4)
        np.testing.assert_almost_equal(p, e_p, 4)


class TestSynthesizeChebyshevNTF:

    def setUp(self):
        pass

    def test_8_4_3(self):
        order = 8
        osr = 4
        h_inf = 3
        z, p, k = synthesizeChebyshevNTF(order, osr, 1, h_inf)
        e_k = 1
        e_z = [716.675171843707e-003 + 697.407125044473e-003j,
               716.675171843707e-003 - 697.407125044473e-003j,
               987.024523733349e-003 + 160.569578528934e-003j,
               987.024523733349e-003 - 160.569578528934e-003j,
               787.924470008031e-003 + 615.771897347195e-003j,
               787.924470008031e-003 - 615.771897347195e-003j,
               899.412094416429e-003 + 437.101686587289e-003j,
               899.412094416429e-003 - 437.101686587289e-003j]
        e_p = [626.054825241768e-003 + 122.563529859845e-003j,
               626.054825241768e-003 - 122.563529859845e-003j,
               603.812988962951e-003 + 354.087262468194e-003j,
               603.812988962951e-003 - 354.087262468194e-003j,
               583.053683660931e-003 + 552.205015613472e-003j,
               583.053683660931e-003 - 552.205015613472e-003j,
               599.974045233308e-003 + 710.003702019636e-003j,
               599.974045233308e-003 - 710.003702019636e-003j]
        e_z = cplxpair(e_z)
        e_p = cplxpair(e_p)
        np.testing.assert_almost_equal(k, e_k, 4)
        np.testing.assert_almost_equal(z, e_z, 4)
        np.testing.assert_almost_equal(p, e_p, 4)


class TestClans:
    def setUp(self):
        pass

    def test_5_32_5_095_1(self):
        z, p, k = clans(5, 32, 5, 0.95, 1)
        e_k = 1
        e_z = [1.00000000000000e+000,
               998.603018798634e-003 + 52.8394818885978e-003j,
               998.603018798634e-003 - 52.8394818885978e-003j,
               996.045312843931e-003 + 88.8466924631155e-003j,
               996.045312843931e-003 - 88.8466924631155e-003j]
        e_p = [418.352260015271e-003,
               489.222137893601e-003 + 170.971685170035e-003j,
               489.222137893601e-003 - 170.971685170035e-003j,
               652.448968088646e-003 + 381.722332008184e-003j,
               652.448968088646e-003 - 381.722332008184e-003j]
        e_z = cplxpair(e_z)
        e_p = cplxpair(e_p)
        z = cplxpair(z)
        p = cplxpair(p)
        np.testing.assert_almost_equal(k, e_k, 4)
        np.testing.assert_almost_equal(z, e_z, 4)
        np.testing.assert_almost_equal(p, e_p, 4)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
import cvxopt as opt
from ..defs import *
from ..scalars import cvxpy_obj

# Class definition
class cvxpy_semidefinite_cone(object):
    r"""
    | :math:`\{X \in \mathbb{R}^{n \\times n} \ |
      \ X = X^T, \ X \succeq 0 \} = \mathbb{S}_+^n`.
    """

    # Method: __init__
    def __init__(self):
        """
        Class constructor.
        """

        self.type = SET
        self.name = 'semidefinite_cone'
        self.expansion_type = SDC

    # Method: valid_shape
    def valid_shape(self,shape):

        return shape[0] == shape[1]

    # Method: __str__
    def __str__(self):

        return self.name

    # Method: _construct
    def _construct(self,el,mp,n):

        m = int(el.shape[0])
        G = opt.spmatrix(0.0,[],[],(m*m,n))
        h = opt.matrix(0.0,(m*m,1))
        for j in range(0,m,1):
            for i in range(0,m,1):
                if np.isscalar(el[i,j]):
                    h[j*m+i,0] = el[i,j]*1.
                elif type(el[i,j]) is cvxpy_obj:
                    h[j*m+i,0] = el[i,j].value*1.
                else:
                    G[j*m+i,mp[el[i,j]]] = -1.
        return G,h,m

# Create instance
semidefinite_cone = cvxpy_semidefinite_cone()



# -*- coding: utf-8 -*-

# Copyright (c) 2013, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Acoustic weighting functions (:mod:`pydsm.audio_weightings`)
============================================================

Some standard acoustic weighting functions.

This module includes the A-, B- and C-weightings from the
ANSI Standards S1.4-1983 and S1.42-2001.

It also includes the D-weighting from the now withdrawn IEC 537.

It also includes the F-weighting proposed by R. A. Wannamaker.

The weighting functions can be expressed either in terms of
acoustic power or in terms of signal amplitude.

The weighting functions are also available in terms of a filter-based
implementation. In this case, be careful since no normalization is
present so that the gain at 1 kHz can be arbitrary. The filter
transfer function is referred to a signal amplitude weighting.

.. currentmodule:: pydsm.audio_weightings


Weighting functions
-------------------

.. autosummary::
   :toctree: generated/

    a_weighting
    b_weighting
    c_weighting
    d_weighting
    f_weighting


Filter implementation of weighting functions
--------------------------------------------

.. autodata:: a_zpk
   :annotation:
.. autodata:: b_zpk
   :annotation:
.. autodata:: c_zpk
   :annotation:
.. autodata:: d_zpk
   :annotation:
.. autodata:: f_zpk
   :annotation:


Normalization constants
-----------------------

.. autodata:: a_weighting_gain
   :annotation:
.. autodata:: b_weighting_gain
   :annotation:
.. autodata:: c_weighting_gain
   :annotation:
.. autodata:: d_weighting_gain
   :annotation:
.. autodata:: f_weighting_gain
   :annotation:


Notes
-----
The ANSI and IEC weightings are also described in Wikipedia [1]_
and summarized in some illustrative web pages such as [2]_ and
[3]_. The F-weighting is documented in [4]_.

The filter-based implementation of the F-weighting is so high-order that
evaluation of the transfer function may require special care.

.. [1] Wikipedia (https://en.wikipedia.org/wiki/A-weighting)
.. [2] Cross spectrum (https://www.cross-spectrum.com/audio/weighting.html)
.. [3] Product Technology Partners ltd "Noise Measurement Briefing"
   (https://web.archive.org/web/20130225/http://www.ptpart.co.uk/noise-measurement-briefing/)
.. [4] Robert A. Wannamaker "Psychoacoustically Optimal Noise
   Shaping," J. Audio Eng. Soc., Vol. 40 No. 7/8 1992 July/August
"""

from __future__ import division, print_function

import numpy as np

__all__ = ["a_zpk", "a_weighting", "b_zpk", "b_weighting",
           "c_zpk", "c_weighting", "d_zpk", "d_weighting",
           "f_zpk", "f_weighting"]

a_zpk = (2*np.pi*np.asarray([0., 0., 0., 0.]),
         2*np.pi*np.asarray([-20.6, -20.6, -107.7, -739.9, -12200., -12200.]),
         (2*np.pi*12200.)**2)
"""A-weighting filter in zpk form."""

b_zpk = (2*np.pi*np.asarray([0., 0., 0.]),
         2*np.pi*np.asarray([-20.6, -20.6, -158.5, -12200., -12200.]),
         (2*np.pi*12200.)**2)
"""B-weighting filter in zpk form."""

c_zpk = (2*np.pi*np.asarray([0., 0.]),
         2*np.pi*np.asarray([-20.6, -20.6, -12200., -12200.]),
         (2*np.pi*12200.)**2)
"""C-weighting filter in zpk form."""

d_zpk = (2*np.pi*np.asarray([0., -519.8+876.2j, -519.8-876.2j]),
         2*np.pi*np.asarray([-282.7, -1160., -1712+2628j, -1712-2628j]),
         91104.32)
"""D-weighting filter in zpk form."""

f_zpk = (2*np.pi*np.asarray([0., 0., 0.,
                             -580+1030j, -580-1030j,
                             -3180+8750j, -3180-8750j,
                             -3180+8750j, -3180-8750j,
                             -3180+8750j, -3180-8750j]),
         2*np.pi*np.asarray([-180., -180., -180.,
                             -1630., -1630.,
                             -2510+3850j, -2510-3850j,
                             -2510+3850j, -2510-3850j,
                             -2510+3850j, -2510-3850j,
                             -2510+3850j, -2510-3850j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j,
                             -6620+14290j, -6620-14290j]),
         1.6810544531883432e+207)
"""F-weighting filter in zpk form."""

# Note: evaluating the transfer function of f_zpk may require special care
# since the high order implies that for many frequency values both the
# numerator and the denominator take very large values (in magnitude). Taking
# the ratio of large complex values may lead to overflow in numpy even if
# individually the numerator, the denominator and the result should not
# overflow.


def a_weighting(f, normal=True, power=True):
    """Returns the A-weighting as a function of frequency.

    Parameters
    ----------
    f : float or array of floats
        frequency where the weighting function is computed
    normal : bool
        whether the function should be normalized to have unit gain at
        1 kHz.
    power : bool
        whether the function should express the weighting in terms of
        acoustic power or signal amplitude

    Returns
    -------
    w : float or array of floats
        value of the weigting function
    """
    if power:
        return a_weighting(f, normal, power=False)**2
    w = (12200.0**2*f**4)/((f**2+20.6**2) *
                           np.sqrt((f**2+107.7**2) *
                           (f**2+737.9**2))*(f**2+12200.0**2))
    return w if not normal else w*a_weighting_gain

a_weighting_gain = 1/a_weighting(1000, normal=False, power=False)
"""Normalization gain to apply to A-weighting filter (namely, the
   attenuation of the filter at 1 kHz)"""


def b_weighting(f, normal=True, power=True):
    """Returns the B-weighting as a function of frequency.

    Parameters
    ----------
    f : float or array of floats
        frequency where the weighting function is computed
    normal : bool
        whether the function should be normalized to have unit gain at
        1 kHz.
    power : bool
        whether the function should express the weighting in terms of
        acoustic power or signal amplitude

    Returns
    -------
    w : float or array of floats
        value of the weigting function
    """
    if power:
        return b_weighting(f, normal, power=False)**2
    w = (12200.0**2*f**3)/((f**2+20.6**2) *
                           np.sqrt(f**2+158.5**2)*(f**2+12200.0**2))
    return w if not normal else w*b_weighting_gain

b_weighting_gain = 1/b_weighting(1000, normal=False, power=False)
"""Normalization gain to apply to B-weighting filter (namely, the
   attenuation of the filter at 1 kHz)"""


def c_weighting(f, normal=True, power=True):
    """Returns the C-weighting as a function of frequency.

    Parameters
    ----------
    f : float or array of floats
        frequency where the weighting function is computed
    normal : bool
        whether the function should be normalized to have unit gain at
        1 kHz.
    power : bool
        whether the function should express the weighting in terms of
        acoustic power or signal amplitude

    Returns
    -------
    w : float or array of floats
        value of the weigting function
    """
    if power:
        return c_weighting(f, normal, power=False)**2
    w = (12200.0**2*f**2)/((f**2+20.6**2)*(f**2+12200.0**2))
    return w if not normal else w*c_weighting_gain

c_weighting_gain = 1/c_weighting(1000, normal=False, power=False)
"""Normalization gain to apply to C-weighting filter (namely, the
   attenuation of the filter at 1 kHz)"""


def d_weighting(f, normal=True, power=True):
    """Returns the D-weighting as a function of frequency.

    Parameters
    ----------
    f : float or array of floats
        frequency where the weighting function is computed
    normal : bool
        whether the function should be normalized to have unit gain at
        1 kHz. This parameter is ignored, since this weighting function
        is always normalized.
    power : bool
        whether the function should express the weighting in terms of
        acoustic power or signal amplitude

    Returns
    -------
    w : float or array of floats
        value of the weigting function
    """
    if power:
        return d_weighting(f, normal, power=False)**2

    def h(f):
        return (((1037918.48-f**2)**2+1080768.16*f**2) /
                ((9837328.0-f**2)**2+11723776.0*f**2))

    return (f/6.8966888496476E-5 *
            np.sqrt(h(f)/((f**2+79919.29)*(f**2+1345600.0))))

d_weighting_gain = 1.
"""Normalization gain to apply to D-weighting filter (namely, the
   attenuation of the filter at 1 kHz)"""


def f_weighting(f, normal=True, power=True):
    """Returns the F-weighting as a function of frequency.

    Parameters
    ----------
    f : float or array of floats
        frequency where the weighting function is computed
    normal : bool
        whether the function should be normalized to have unit gain at
        1 kHz.
    power : bool
        whether the function should express the weighting in terms of
        acoustic power or signal amplitude

    Returns
    -------
    w : float or array of floats
        value of the weigting function

    Notes
    -----
    The F-weighting function is documented in [1]_.

    .. [1] Robert A. Wannamaker "Psychoacoustically Optimal Noise Shaping,"
       J. Audio Eng. Soc., Vol. 40 No. 7/8 1992 July/August
    """
    if not power:
        return np.sqrt(f_weighting(f, normal, power=True))
    fx = f/1000.
    g = 2.536e-5
    z1 = fx**2
    z2 = ((0.58**2)+(1.03**2)-z1)**2 + 4.0*(0.58**2)*z1
    z3 = ((3.18**2)+(8.75**2)-z1)**2 + 4.0*(3.18**2)*z1
    p1 = 0.18**2+z1
    p2 = 1.63**2+z1
    p3 = ((2.51**2)+(3.85**2)-z1)**2 + 4.0*(2.51**2)*z1
    p4 = ((6.62**2)+(14.29**2)-z1)**2 + 4.0*(6.62**2)*z1
    w = ((g*((z1**3)*z2*(z3**3)) /
         ((p1**3)*(p2**2)*(p3**4))*((1e5/p4)**20)))
    return w if not normal else w*f_weighting_gain

# Set normalization gain
f_weighting_gain = 1/f_weighting(1000, normal=False, power=True)
"""Normalization gain to apply to F-weighting filter (namely, the
   attenuation of the filter at 1 kHz)"""



# -*- coding: utf-8 -*-

# Copyright (c) 2015–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes portions of code ported from the "Frequency-Domain
# Min-Max Optimization for Delta-Sigma Modulators" Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/36187)
# covered by the following copyright and permission notice
#
# Copyright (c) 2012 Masaaki Nagahara
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from __future__ import division, print_function

import numpy as np
import cvxpy


def ntf_fir_from_digested(order, osrs, H_inf, f0s, zf, **opts):
    """
    Synthesize FIR NTF with minmax approach from predigested specification

    Version for the cvxpy modeler.
    """
    verbose = opts['show_progress']
    if opts['cvxpy_opts']['solver'] == 'cvxopt':
        opts['cvxpy_opts']['solver'] = cvxpy.CVXOPT
    elif opts['cvxpy_opts']['solver'] == 'scs':
        opts['cvxpy_opts']['solver'] = cvxpy.SCS

    # State space representation of NTF
    A = np.eye(order, order, 1)
    B = np.vstack((np.zeros((order-1, 1)), 1.))
    # C contains the NTF coefficients
    D = np.array([[1.]])

    # Set up the problem
    bands = len(f0s)
    c = cvxpy.Variable((1, order))
    F = []
    gg = cvxpy.Variable((bands, 1))

    for idx in range(bands):
        f0 = f0s[idx]
        osr = osrs[idx]
        omega0 = 2*f0*np.pi
        Omega = 1./osr*np.pi
        P = cvxpy.Variable((order, order), symmetric=True)
        Q = cvxpy.Variable((order, order), PSD=True)
        if f0 == 0:
            # Lowpass modulator
            M1 = (A.T @ P @ A + Q @ A +
                  A.T @ Q - P - 2*Q*np.cos(Omega))
            M2 = A.T @ P @ B + Q @ B
            M3 = B.T @ P @ B - gg[idx, 0]
            M = cvxpy.bmat([[M1, M2, c.T],
                            [M2.T, M3, D],
                            [c, D, np.array([[-1]])]])
            F += [M << 0]
            if zf:
                # Force a zero at DC
                F += [cvxpy.sum_entries(c) == -1]
        else:
            # Bandpass modulator
            M1r = (A.T @ P @ A + Q @ A*np.cos(omega0) +
                   A.T @ Q*np.cos(omega0) -
                   P - 2*Q*np.cos(Omega))
            M2r = A.T @ P @ B + Q @ B*np.cos(omega0)
            M3r = B.T @ P @ B - gg[idx, 0]
            M1i = A.T @ Q*np.sin(omega0) - Q @ A*np.sin(omega0)
            M21i = -Q @ B*np.sin(omega0)
            M22i = B.T @ Q*np.sin(omega0)
            Mr = cvxpy.bmat([[M1r, M2r, c.T],
                             [M2r.T, M3r, D],
                             [c, D, np.array([[-1]])]])
            Mi = cvxpy.bmat([[M1i, M21i, np.zeros((order, 1))],
                             [M22i, np.array([[0]]), np.array([[0]])],
                             [np.zeros((1, order)), np.array([[0]]),
                              np.array([[0]])]])
            M = cvxpy.bmat([[Mr, Mi],
                            [-Mi, Mr]])
            F += [M << 0]
            if zf:
                # Force a zero at z=np.exp(1j*omega0)
                nn = np.arange(order).reshape((order, 1))
                vr = np.matrix(np.cos(omega0*nn))
                vi = np.matrix(np.sin(omega0*nn))
                vn = np.matrix(
                    [-np.cos(omega0*order), -np.sin(omega0*order)])
                F += [c*cvxpy.hstack(vr, vi) == vn]
    if H_inf < np.inf:
        # Enforce the Lee constraint
        R = cvxpy.Variable((order, order), PSD=True)
        MM = cvxpy.bmat([[A.T @ R @ A - R, A.T @ R @ B, c.T],
                         [B.T @ R @ A, -H_inf**2 + B.T @ R @ B, D],
                         [c, D, np.array([[-1]])]])
        F += [MM << 0]
    target = cvxpy.Minimize(cvxpy.max(gg))
    p = cvxpy.Problem(target, F)
    p.solve(verbose=verbose, **opts['cvxpy_opts'])
    return np.hstack((1, np.asarray(c.value)[0, ::-1]))
